/* a module to convert a US NTSC TiVo to PAL input and output

  tridge@samba.org, 2000

	Updated By Andrew Palm 2002
	Updated to releax all Australian Channels
*/

#include <linux/config.h>
#include <linux/module.h>
#include <linux/fs.h>
#include <asm/ppcops.h>
#include "aus_chans.h"

#define VERSION				"2.5.1"
#define MODULE_HEADER_LENGTH		0x4C
#define TUNER_BUSADDR			0xC2

/* offsets into pxmpegdecode.o*/
#define offset_cs22NTSCConfig_val	0x1D3A
#define offset_cs22_init		0x5668

/* offsets into i2c.o*/
#define offset_I2CRead			0x0000
#define offset_I2CWrite			0x007C
#define offset_I2CWriteNoSub		0x0168
#define offset_I2CWriteMsp		0x039C

/* offsets into fpga7114.o */
#define offset_standin_tune		0x1FE4
#define offset_fpga_hw_ioctl		0x28A8
#define offset_BangRegs			0x3A6C
#define offset_getFreq_C2		0x7027
#define offset_getFreq_C6		0x7037

static u32 base_i2c = 0;
static u32 base_fpga7114 = 0;
static u32 base_pxmpegdecode = 0;

void base_address_ref(void)
{
/* IMPORTANT: leave this dummy proc at the top of this file.
   Its code *must* reside in the first 4K of the module.
 */
}

/* useful to know what channel we are on */
static int channel = -1;

/* write one byte to a i2c module */
static int I2CWriteByte(int module, int subaddress, unsigned char v)
{
  int (*I2CWrite)(int, int, unsigned char *, int) = (int (*)(int, int, unsigned char *, int)) (base_i2c + offset_I2CWrite);

  return I2CWrite(module, subaddress, &v, 1);
}

/* read one byte to a i2c module */
static unsigned char iicr1(int module, int subaddress)
{
  unsigned char v;
  int (*I2CRead)(int, int, unsigned char *, int) = (int (*)(int, int, unsigned char *, int)) (base_i2c + offset_I2CRead);
  
  I2CRead(module, subaddress, &v, 1);
  return v;
}

static void cache_flush(u32 addr)
{
  __asm ("dcbf 0,%0\n\t"
         "sync \n\t"
         "icbi 0,%0 \n\t"
         "isync \n\t"
         : : "r" (addr));
}

/* write to a CXD1922 register. The CXD is mapped into virtual memory at 0xfc000000 */
static void CxdWrite(u32 addr, u16 value)
{
  u16 *dest = (u16 *)(0xfc000000 | addr);
  *dest = value;
}

/* a hook for the fn in i2c.o that writes to a i2c module that has no subaddess (ie. the tuner) useful for debugging */
static int iicWriteNoSubAddr_x(int module, unsigned char *data, int len)
{
  int (*I2CWriteNoSub)(int, unsigned char *, int) = (int (*)(int, unsigned char *, int)) (base_i2c + offset_I2CWriteNoSub);
  int ret;

  ret = I2CWriteNoSub(module, data, 4);
  printk("PALMOD: I2CWriteNoSub(0x%x, %p, 4) -> %d\n", module, data, ret);
  return ret;
}

/* a hook for i2c read fn in i2c.o */
static int iicRead_x(int module, int subaddress, unsigned char *data, int len)
{
  int ret;
  int (*I2CRead)(int, int, unsigned char *, int) = (int (*)(int, int, unsigned char *, int)) (base_i2c + offset_I2CRead);

  ret = I2CRead(module, subaddress, data, len);
//  printk("PALMOD: I2CRead(0x%x, 0x%x, %p, %d) -> %d\n", module, subaddress, data, len, ret);
  return ret;
}

/* a hook for the basic i2c write fn */
static int iicWrite_x(int module, int subaddress, unsigned char *data, int len)
{
  int (*I2CWrite)(int, int, unsigned char *, int) = (int (*)(int, int, unsigned char *, int)) (base_i2c + offset_I2CWrite);

  int ret=0;
  // i2c bus 0x88 holds the MSP audio chip
  if (module == 0x88) {
    u16 *v = (u16 *)data;
//    printk("PALMOD: I2CWriteMsp(0x%x, 0x%x, 0x%x)\n", subaddress, v[0], v[1]);
    // if we are on a broadcast channel then we need to force the audio to come from the right pin
    if (channel >= 0 && v[0] == 0x13) {
      // force the tuner audio - attached to scart1
      v[1] = 0;
    }
  }

  // if we are on the tuner then force the tuner input in the 7114
  if (module == 0x42 && subaddress == 2 && channel >= 0) {
    return I2CWriteByte(0x42, 0x2, 0x80);
  }
  ret = I2CWrite(module, subaddress, data, len);
//  printk("PALMOD: I2CWrite(0x%x, 0x%x, %p, %d) -> %d\n", module, subaddress, data, len, ret);
  return ret;
}

static int standin_tune_x(u32 a1, u32 a2, u32 a3)
{
  int ret;
  int (*standin_tune)(u32,u32,u32) = (int (*)(u32,u32,u32)) ( base_fpga7114 + offset_standin_tune );


  printk("PALMOD: channel=%d\n", *((u32 *)(a1+16)));

  *((int *)(a1+16)) = -1;

  ret = standin_tune(a1, a2, a3);

  printk("PALMOD: standin_tune(%x, %x, %x) -> %d\n", a1, *(u32 *)a2, a3, ret);

  return ret;
}

static void set_tuner(int channel)
{
  int (*I2CWrite)(int, int, unsigned char *, int) = (int (*)(int, int, unsigned char *, int)) (base_i2c + offset_I2CWrite);
  int (*I2CWriteNoSub)(int, unsigned char *, int) = (int (*)(int, unsigned char *, int)) (base_i2c + offset_I2CWriteNoSub);

  if (channel > MAX_CHANS) return;

  if (tv_chans[channel].chan) {
    /* we attach to pins 56 and 57 of the MSP */
    u32 mspv = 0x00130000;
    I2CWriteByte(0x42, 0x2, 0x80);
    I2CWriteByte(0x42, 0xC, 0x7F);	// Fix washed-out colours - wkt
    I2CWrite(0x88, 0x12, (u8 *) &mspv, 4);
    // printk("PALMOD: tuned to 0x%x\n", tv_chans[channel].chan);
    I2CWriteNoSub(TUNER_BUSADDR, (u8 *) &tv_chans[channel].chan, 4);
   }
}

static int fpga_hw_ioctl_x(u32 a1, u32 a2, u32 a3, u32 a4)
{
  int (*fpga_hw_ioctl)(u32,u32,u32,u32) = (int (*)(u32,u32,u32,u32)) ( base_fpga7114 + offset_fpga_hw_ioctl );
  int ret;

  if (a3 == 0x8000) {
    u32 *p = (u32 *)a4;
    channel = p[1];
    printk("PALMOD: 8000: %x %x %x %x\n", p[0], p[1], p[2], p[3]);
    printk("PALMOD: channel=%d\n", channel);

    // The 0x220 sets the screen resolution to be 544 pixels deep.
    // For XSVCD PAL resolution (480x576), you might want to change
    // this to 576.

    p[1] = ~0;
    p[3] = 0x220;
  }

  ret = fpga_hw_ioctl(a1, a2, a3, a4);

  if (channel >= 0) set_tuner(channel);

//  printk("PALMOD: fpga_hw_ioctl(%x, %x, %x, %x) -> %d\n", a1, a2, a3, a4, ret);

  return ret;
}

static void patch_cs22(int pal)
{
  u16 *p = (u16 *)(base_pxmpegdecode + offset_cs22NTSCConfig_val);
  *p = pal?8192:0;
  cache_flush((u32)p);
}

static void patch_i2c(u8 v)
{
  u8 *p;

  p = (u8 *) (base_fpga7114 + offset_getFreq_C6);
  *p = v;

  cache_flush((u32)p);
}

static void video_setup(void)
{
  int (*cs22_init)(void) = (int (*)(void)) (base_pxmpegdecode + offset_cs22_init);
  int pal=1;
  static int last_pal = 0;

  /* auto-detect PAL or NTSC */
  if ((iicr1(0x42, 0x1f) & 0x60) == 0x20) {
    pal = 0;
  }

  patch_cs22(pal);

  patch_i2c(TUNER_BUSADDR);

  if (last_pal != pal) {
      cs22_init();
  }

  last_pal = pal;

  if (pal) {
    CxdWrite(0x1e0, 1);
    // CxdWrite(0x204, 1);
    CxdWrite(0x222, -16);
    CxdWrite(0x224, 2);
  } else {
    CxdWrite(0x1e0, 0);
    // CxdWrite(0x204, 0);
    CxdWrite(0x222, 0);
    CxdWrite(0x224, 0);
  }

  if (pal) {
    I2CWriteByte(0x42, 0x08, 0x98);
    I2CWriteByte(0x42, 0x03, 0x33);
    I2CWriteByte(0x42, 0x04, 0x49);
    I2CWriteByte(0x42, 0x05, 0x49);
    I2CWriteByte(0x42, 0x09, 0x00);
    I2CWriteByte(0x42, 0x0e, 0x04);
    I2CWriteByte(0x42, 0x0f, 0x00);
    I2CWriteByte(0x42, 0x86, 0x40);
    I2CWriteByte(0x42, 0x9a, 0x22);
    I2CWriteByte(0x42, 0x9b, 0x1);
  } else {
    I2CWriteByte(0x42, 0x08, 0xc8);
    I2CWriteByte(0x42, 0x03, 0x23);
    I2CWriteByte(0x42, 0x09, 0x40);
    I2CWriteByte(0x42, 0x0e, 0x1);
    I2CWriteByte(0x42, 0x86, 0xc0);
    I2CWriteByte(0x42, 0x9a, 0xf2);
    I2CWriteByte(0x42, 0x9b, 0x0);
  }

  if (pal) {
    I2CWriteByte(0x8c, 0x61, 0x82);
    I2CWriteByte(0x8c, 0x63, 0xcb);
    I2CWriteByte(0x8c, 0x64, 0x8a);
    I2CWriteByte(0x8c, 0x65, 0x09);
    I2CWriteByte(0x8c, 0x66, 0x2a);
    I2CWriteByte(0x8c, 0x6e, 0x80);
    I2CWriteByte(0x8c, 0x7a, 0x14);
    I2CWriteByte(0x8c, 0x7b, 0x35);
  } else {
    I2CWriteByte(0x8c, 0x61, 0x80);
    I2CWriteByte(0x8c, 0x63, 0x1f);
    I2CWriteByte(0x8c, 0x64, 0x7c);
    I2CWriteByte(0x8c, 0x65, 0xf0);
    I2CWriteByte(0x8c, 0x66, 0x21);
    I2CWriteByte(0x8c, 0x6e, 0x80);
    I2CWriteByte(0x8c, 0x7a, 0x14);
    I2CWriteByte(0x8c, 0x7b, 0x35);
  }

  printk("PALMOD: video setup for %s\n", pal?"PAL":"NTSC");
}

static int getFreq_x(int chan, int u1, u32 *p1, u32 *p2)
{
  unsigned char v;
  int (*I2CRead)(int, int, unsigned char *, int) = (int (*)(int, int, unsigned char *, int)) (base_i2c + offset_I2CRead);
  u32 f=0;

  printk("PALMOD: getFreq chan=%d\n", chan);

  if (chan > MAX_CHANS) return 0;

  if (tv_chans[chan].freq) f = (tv_chans[chan].freq + 38900) * 1000;

  I2CRead(0x42, 0x1f, &v, 1);

  *p1 = f;
  if (tv_chans[chan].freq > 385250) {
    *p2 = 2;
  } else if (tv_chans[chan].freq > 137250) {
    *p2 = 1;
  } else {
    *p2 = 0;
  }

  return 0;
}

static int BangRegs_x(int a1, int a2, int a3)
{
  int ret;

  int (*BangRegs)(int,int,int) = (int (*)(int,int,int)) ( base_fpga7114 + offset_BangRegs );

  ret = BangRegs(a1, a2, a3);
//  printk("PALMOD: BangRegs(%x, %x, %x) -> %d\n", a1, a2, a3, ret);
  video_setup();

  return ret;
}

static void install_fpga_bypass(u32 newfn, u32 call)
{
  u32 old;
  call += base_fpga7114;
  old = *(u32 *)call;
  *(u32 *)call = (0x48000001) | ((newfn - call)&0xFFFFFF);
  cache_flush(call);
  printk("PALMOD: patched at 0x%x with 0x%x (was 0x%x)\n", call, *(u32 *)call, old);
}

int init_module(void)
{
  int count=0;
  /* get our base address */
  struct module *mod_base = (struct module *)(((int)base_address_ref)&0xFFFFF000);
  /* walk the structure and search for other modules */
  while (mod_base->next){
    if (strcmp("pxmpegdecode", mod_base->name) == 0) {
      base_pxmpegdecode = (u32)mod_base + MODULE_HEADER_LENGTH;
      printk("PALMOD: found %s at %08x\n", mod_base->name, base_pxmpegdecode);
      count++;
    }
    if (strcmp("fpga7114", mod_base->name) == 0) {
      base_fpga7114 = (u32)mod_base + MODULE_HEADER_LENGTH;
      printk("PALMOD: found %s at %08x\n", mod_base->name, base_fpga7114);
      count++;
    }
    if (strcmp("i2c", mod_base->name) == 0) {
      base_i2c = (u32)mod_base + MODULE_HEADER_LENGTH;
      printk("PALMOD: found %s at %08x\n", mod_base->name, base_i2c);
      count++;
    }
    mod_base = mod_base->next;
  }

  if (count == 3) {
    install_fpga_bypass((u32)BangRegs_x, 0x48B0);
    install_fpga_bypass((u32)BangRegs_x, 0x5178);

    install_fpga_bypass((u32)getFreq_x, 0x708C);

    install_fpga_bypass((u32)iicWriteNoSubAddr_x, 0x6c84);

    install_fpga_bypass((u32)iicRead_x, 0x5D74);
    install_fpga_bypass((u32)iicRead_x, 0x5E48);
    install_fpga_bypass((u32)iicRead_x, 0x5ED4);
    install_fpga_bypass((u32)iicRead_x, 0x5FF4);
    install_fpga_bypass((u32)iicRead_x, 0x6064);

    {
      u32 addrs[] = {
        0x523c, 0x5338, 0x534c, 0x537c, 0x5390, 0x53b8, 0x541c, 0x54a0,
        0x54b4, 0x54c8, 0x54ec, 0x550c, 0x5530, 0x5548, 0x55cc, 0x55e0,
        0x55f4, 0x5618, 0x5638, 0x565c, 0x5678, 0x56f8, 0x570c, 0x5720,
        0x5744, 0x5764, 0x57c8, 0x57e0, 0x57f8, 0x581c, 0x5838, 0x5858,
        0x588c, 0x59c8, 0x5b48, 0x5b88, 0x5c60, 0x5cb0, 0x5d08, 0x5e68,
        0x5e88, 0x5ef4, 0x6018, 0x609c, 0x6260, 0x62d8, 0x6308, 0x6334,
        0x63b8, 0x63d4, 0x6458, 0x6488, 0x64b4, 0x651c, 0x654c, 0x6578,
        0x65c0, 0x65f0, 0
      };
      int i;
      for (i=0; addrs[i]; i++) {
        install_fpga_bypass((u32)iicWrite_x, addrs[i]);
      }
    }

    install_fpga_bypass((u32)standin_tune_x, 0x2D6C);
    install_fpga_bypass((u32)standin_tune_x, 0x2D8C);

    install_fpga_bypass((u32)fpga_hw_ioctl_x, 0x4c);

    video_setup();

    printk("PALMOD: version %s - bypass installed\n", VERSION);
  } else {
    printk("PALMOD: version %s - bypass NOT installed (failed to get required module's base address).\n", VERSION);
  }

  return 0;
}

void cleanup_module(void)
{
  printk("PALMOD: version %s - Cannot be removed!\n", VERSION);
}
